/* 
   (c) Copyright 2001-2002  convergence integrated media GmbH.
   (c) Copyright 2002-2005  convergence GmbH.

   All rights reserved.

   Written by Denis Oliver Kropp <dok@directfb.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread.h>
#include <directfb.h>

#include <direct/debug.h>
#include <direct/mem.h>
#include <direct/messages.h>

#include "util.h"

#include "box.h"
#include "window.h"

D_DEBUG_DOMAIN(LiteBoxDomain, "LiTE/Box", "LiteBox");

static void draw_box_and_children(LiteBox *box, const DFBRegion *region, 
                                   DFBBoolean clear);
static void defocus_me_or_children (LiteBox *box);
static void deenter_me_or_children (LiteBox *box);

DFBResult 
lite_new_box(LiteBox **box, 
             LiteBox  *parent, 
             int       x, 
             int       y, 
             int       width, 
             int       height)
{
     LiteBox   *bx = NULL;
     DFBResult  res; 

     LITE_NULL_PARAMETER_CHECK(parent);

     bx = D_CALLOC(1, sizeof(LiteBox));

     bx->parent = parent;

     bx->rect.x = x;
     bx->rect.y = y;
     bx->rect.w = width;
     bx->rect.h = height;

     res = lite_init_box(bx);
     if (res != DFB_OK) {
          D_FREE(bx);
          return res;
     }

     *box = bx;

     D_DEBUG_AT(LiteBoxDomain, "Created new box object: %p\n", bx);

     return DFB_OK;
}

DFBResult 
lite_draw_box(LiteBox         *box, 
              const DFBRegion *region, 
              DFBBoolean       flip)
{
     LITE_NULL_PARAMETER_CHECK(box);

     if (region) {
          D_DEBUG_AT(LiteBoxDomain, 
                   "Update box: %p at %d,%d %dx%d\n", box, region->x1, region->y1,
                   region->x2 - region->x1 + 1, region->y2 - region->y1 + 1);

          draw_box_and_children(box, region, DFB_TRUE);

          if (flip)
               box->surface->Flip(box->surface, region, 0);
     }
     else {
          DFBRegion reg = { 0, 0, box->rect.x + box->rect.w - 1, 
                              box->rect.y + box->rect.h - 1 };

          lite_draw_box(box, &reg, flip);
     }

     if (box->type == LITE_TYPE_WINDOW) {
         LiteWindow* window = (LiteWindow*)box;
         window->flags |= LITE_WINDOW_DRAWN;
         window->window->SetOpacity(window->window, window->opacity);
     }

     return DFB_OK;
}

DFBResult 
lite_update_box(LiteBox *box, const DFBRegion *region)
{
     LITE_NULL_PARAMETER_CHECK(box);

     /* don't update region that's part of an invisible box */
     if (box->is_visible == 0)
          return DFB_OK;

     if (region) {
          if (region->x2 < region->x1 || region->y2 < region->y1)
               return DFB_OK;

          if (region->x1 > box->rect.w - 1 || region->x2 < 0 ||
              region->y1 > box->rect.h - 1 || region->y2 < 0)
               return DFB_OK;

          if (box->parent) {
               DFBRegion trans;

               trans.x1 = region->x1 + box->rect.x;
               trans.y1 = region->y1 + box->rect.y;
               trans.x2 = region->x2 + box->rect.x;
               trans.y2 = region->y2 + box->rect.y;

               lite_update_box(box->parent, &trans);
          }
          else if (box->type == LITE_TYPE_WINDOW)
               lite_update_window(LITE_WINDOW(box), region);
          else
               D_DEBUG_AT(LiteBoxDomain,
                    "Can't update a box without a top level parent!\n");
     }
     else {
          DFBRegion reg = { 0, 0, box->rect.w - 1, box->rect.h - 1 };

          lite_update_box(box, &reg);
     }

     return DFB_OK;
}

DFBResult 
lite_destroy_box(LiteBox *box)
{
     int i;

     LITE_NULL_PARAMETER_CHECK(box);

     defocus_me_or_children (box);
     deenter_me_or_children (box);

     /* remove the child from the parent child array, unless 
        it's a window (no parent), or recursive */
     if (box->parent != NULL && box->type != LITE_TYPE_WINDOW) {
            lite_remove_child(box->parent, box);
     }

     /* destroy children */
     for (i=0; i<box->n_children; i++) {
          box->children[i]->parent = NULL;
          box->children[i]->Destroy(box->children[i]);
     }

     /* free surface and used memory */
     box->surface->Release(box->surface);

     if (box->children)
        D_FREE(box->children);

     D_FREE(box);

     box = NULL;

     return DFB_OK;
}


DFBResult 
lite_init_box(LiteBox *box)
{
     DFBResult ret;

     LITE_NULL_PARAMETER_CHECK(box);

     /* Get sub surface */
     ret = box->parent->surface->GetSubSurface(box->parent->surface,
                                                &box->rect, &box->surface);
     if (ret) {
          DirectFBError("lite_init_box: Surface->GetSubSurface", ret);
          return -1;
     }

     box->Destroy = lite_destroy_box;

     lite_add_child(box->parent, box);

     box->handle_keys = 1;      /* by default all liteboxes handle keys */
     box->is_visible  = 1;      /* by default all liteboxes are visible */
     box->is_active = 1;        /* by default all liteboxes are active */ 
     box->type = LITE_TYPE_BOX;

     return ret;
}

DFBResult 
lite_reinit_box_and_children(LiteBox *box)
{
     int i;
     DFBResult ret = DFB_OK;

     LITE_NULL_PARAMETER_CHECK(box);

     /* Get new sub surface */
     if (box->parent) {
          IDirectFBSurface *surface;
          IDirectFBFont *font;

          ret = box->parent->surface->GetSubSurface(box->parent->surface,
                                                     &box->rect, &surface);
          if (ret) {
               DirectFBError("lite_reinit_box: Surface->GetSubSurface", ret);
               return ret;
          }

          /* transfer the font to the new sub-surface before releasing
             the old sub-surface */
          ret = box->surface->GetFont(box->surface, &font);
          if (font)
               ret = surface->SetFont(surface, font); 

          ret = box->surface->Release(box->surface);
          box->surface = surface;
     }

     /* reinit children */
     for (i=0; i<box->n_children; i++)
          ret = lite_reinit_box_and_children(box->children[i]);

    return ret;
}

DFBResult 
lite_clear_box(LiteBox *box, const DFBRegion *region)
{
     LITE_NULL_PARAMETER_CHECK(box);

     if (box->parent) {
          DFBRegion reg;

          if (region) {
               reg = *region;

               reg.x1 += box->rect.x;
               reg.x2 += box->rect.x;
               reg.y1 += box->rect.y;
               reg.y2 += box->rect.y;
          }
          else {
               reg.x1 = box->rect.x;
               reg.y1 = box->rect.y;
               reg.x2 = box->rect.x + box->rect.w - 1;
               reg.y2 = box->rect.y + box->rect.h - 1;
          }

          if (box->parent->Draw)
               box->parent->Draw(box->parent, &reg, DFB_TRUE);
          else
               lite_clear_box(box->parent, &reg);
     }
     else {
          D_DEBUG_AT(LiteBoxDomain, 
               "lite_clear_box: No parent present in the LiteBox\n");
     }

     return DFB_OK;
}

DFBResult
lite_add_child(LiteBox *parent, LiteBox *child)
{
     LiteWindow *window = NULL;

     LITE_NULL_PARAMETER_CHECK(parent);
     LITE_NULL_PARAMETER_CHECK(child);

     /* update the child array */
     parent->n_children++;

     parent->children = D_REALLOC(parent->children,
                              sizeof(LiteBox*) * parent->n_children);

     parent->children[parent->n_children-1] = child;

     /* get a possible window that the box is included into */
     window = lite_find_my_window(parent);

     /* trigger OnBoxAdded callback if the box will be added */
     if (window != NULL) {
        if (window->OnBoxAdded) {
            window->OnBoxAdded(window, child);
        }
     }
     
     return DFB_OK;
}

DFBResult
lite_remove_child(LiteBox *parent, LiteBox *child)
{
     int i;
     LiteWindow *window = NULL;

     LITE_NULL_PARAMETER_CHECK(parent);
     LITE_NULL_PARAMETER_CHECK(child);

     /* get a possible window that the box is included into */
     window = lite_find_my_window(child);

     if (window != NULL) {
        /* triggerOnBoxToBeRemoved in case a box is removed */
        if (window->OnBoxToBeRemoved) {
            window->OnBoxToBeRemoved(window, child);
        }

        /* adjust the focus in case this box has the window focus, if so
           then default it to the window box */
        if (window->focused_box == child) {
             window->focused_box = LITE_BOX(window);
        }

        /* also make sure that if the box is an entered_box,
           then set this to NULL, as it's now removed */
        if (window->entered_box == child) {
             window->entered_box = NULL;
        } 

        /* make sure that if the box is an entered box,
           then set this to NULL as it's now removed */
        if (window->drag_box == child) {
             window->drag_box = NULL;
        }
     }
     
     /* remove the child from the parent array */
     for (i = 0; i < parent->n_children; i++) {
          if (parent->children[i] == child) 
               break;
     }

     /* update the child array */
     if (i == parent->n_children) {
          D_DEBUG_AT(LiteBoxDomain, 
          "Could not find the child in parent's child directory for removal.\n");

          return DFB_FAILURE;
     }

     /* force an update for the area occupied by the child */
     lite_update_box(child, NULL);

     parent->n_children--;
     for (; i < parent->n_children; i++)
          parent->children[i] = parent->children[i+1];

     parent->children = D_REALLOC(parent->children, 
                         sizeof(LiteBox*) * parent->n_children);

     return DFB_OK;
}

/* internals */

static void 
draw_box_and_children(LiteBox *box, const DFBRegion *region, DFBBoolean clear)
{
     int i;

     D_ASSERT(box != NULL);

     if (box->is_visible == 0)
        return;

     if (region->x2 < region->x1 || region->y2 < region->y1)
          return;

     if (region->x1 > box->rect.w - 1 || region->x2 < 0 ||
         region->y1 > box->rect.h - 1 || region->y2 < 0)
          return;

     D_DEBUG_AT(LiteBoxDomain,"Draw box:   %p at %d,%d %dx%d\n",
              box, box->rect.x, box->rect.y, box->rect.w, box->rect.h);

     /* draw box */
     if (box->Draw)
          box->Draw(box, region, clear);

     /* draw children */
     for (i=0; i<box->n_children; i++) {
          DFBRegion  trans;
          LiteBox   *child = box->children[i];

          trans.x1 = region->x1 - child->rect.x;
          trans.y1 = region->y1 - child->rect.y;
          trans.x2 = region->x2 - child->rect.x;
          trans.y2 = region->y2 - child->rect.y;

          draw_box_and_children(child, &trans, DFB_FALSE);
     }
}

static void 
defocus_me_or_children (LiteBox *box)
{
     LiteWindow* top = NULL;
     LiteBox* traverse = NULL;

     D_ASSERT(box != NULL); 
    
     /* find the window to figure out who has focus */
     top = lite_find_my_window (box);

     /* traverse from the focused box to see if box is its ancestor */
     if (NULL != top) {
          traverse = top->focused_box;
          while (traverse) {
               if (traverse == box) {
                    top->focused_box = NULL;
                    break;
               }
               traverse = traverse->parent;
          }
     }
}

static void
deenter_me_or_children (LiteBox *box)
{
     LiteWindow* top = NULL;
     LiteBox* traverse = NULL;

     D_ASSERT(box != NULL);

     /* find the window to figure out who is the entered box */
     top = lite_find_my_window (box);

     /* traverse from the entered box to see if box is its ancestor */
     if (NULL != top) {
          traverse = top->entered_box;
          while (traverse) {
               if (traverse == box) {
                    top->entered_box = NULL;
                    break;
               }
               traverse = traverse->parent;
          }
     }
}

